"""
lanpartydb.deserialization
~~~~~~~~~~~~~~~~~~~~~~~~~~

TOML deserialization to objects

:Copyright: 2024-2025 Jochen Kupperschmidt
:License: MIT
"""

from decimal import Decimal
from pathlib import Path
import tomllib
from typing import Any

from .models import Location, Party, PartyLinks, Resource, Series, SeriesLinks


# series list


def deserialize_series_list_from_toml_file(filename: Path) -> list[Series]:
    """Deserialize list of series from a TOML file."""
    toml = filename.read_text()
    return deserialize_series_list_from_toml(toml)


def deserialize_series_list_from_toml(toml: str) -> list[Series]:
    """Deserialize list of series from a TOML document."""
    data = _load_toml(toml)
    return _deserialize_series_list_from_dict(data)


def _deserialize_series_list_from_dict(data: dict[str, Any]) -> list[Series]:
    """Build list of series from a dictionary."""
    return [
        _deserialize_series_from_dict(series_dict)
        for series_dict in data.get('series', [])
    ]


# series


def deserialize_series_from_toml_file(filename: Path) -> Series:
    """Deserialize series from a TOML file."""
    toml = filename.read_text()
    return deserialize_series_from_toml(toml)


def deserialize_series_from_toml(toml: str) -> Series:
    """Deserialize series from a TOML document."""
    data = _load_toml(toml)
    return _deserialize_series_from_dict(data)


def _deserialize_series_from_dict(series_dict: dict[str, Any]) -> Series:
    """Build series from a dictionary."""
    links_dict = series_dict.pop('links', None)
    if links_dict:
        website_dict = links_dict.pop('website', None)
        if website_dict:
            website = Resource(
                url=website_dict['url'],
                offline=website_dict.get('offline', False),
            )
            series_dict['links'] = SeriesLinks(website=website)

    return Series(**series_dict)


# party


def deserialize_party_from_toml_file(filename: Path) -> Party:
    """Deserialize party from a TOML file."""
    toml = filename.read_text()
    return deserialize_party_from_toml(toml)


def deserialize_party_from_toml(toml: str) -> Party:
    """Deserialize party from a TOML document."""
    data = _load_toml(toml)
    return _deserialize_party_from_dict(data)


def _deserialize_party_from_dict(party_dict: dict[str, Any]) -> Party:
    """Build party from a dictionary."""
    location_dict = party_dict.pop('location', None)
    if location_dict:
        party_dict['location'] = Location(**location_dict)

    links_dict = party_dict.pop('links', None)
    if links_dict:
        website_dict = links_dict.pop('website', None)
        if website_dict:
            website = Resource(
                url=website_dict['url'],
                offline=website_dict.get('offline', False),
            )
            party_dict['links'] = PartyLinks(website=website)

    return Party(**party_dict)


# util


def _load_toml(toml: str) -> str:
    return tomllib.loads(toml, parse_float=Decimal)



"""
:Copyright: 2024-2025 Jochen Kupperschmidt
:License: MIT
"""

from textwrap import dedent

from lanpartydb.models import Series, SeriesLinks, Resource
from lanpartydb.serialization import serialize_series_list_to_toml


def test_serialize_series_list_to_toml():
    series_list = [
        Series(
            slug='gammalan',
            title='GammaLAN',
            alternative_titles=[],
            country_codes=['ca', 'us'],
        ),
        Series(
            slug='deltalan',
            title='DeltaLAN',
            alternative_titles=['Δ LAN', 'Δέλτα LAN'],
            country_codes=['au'],
            links=SeriesLinks(
                website=Resource(
                    url='https://www.deltalan.example/',
                ),
            ),
        ),
    ]

    assert serialize_series_list_to_toml(series_list) == dedent("""\
            [[series]]
            slug = "gammalan"
            title = "GammaLAN"
            country_codes = ["ca", "us"]

            [[series]]
            slug = "deltalan"
            title = "DeltaLAN"
            alternative_titles = ["Δ LAN", "Δέλτα LAN"]
            country_codes = ["au"]

            [series.links.website]
            url = "https://www.deltalan.example/"
            """)



"""
:Copyright: 2024-2025 Jochen Kupperschmidt
:License: MIT
"""

import pytest

from lanpartydb.deserialization import deserialize_series_from_toml
from lanpartydb.models import Series, SeriesLinks, Resource


@pytest.mark.parametrize(
    ('toml', 'expected'),
    [
        (
            """
            slug = "megalan"
            title = "MegaLAN"
            """,
            Series(
                slug='megalan',
                title='MegaLAN',
                alternative_titles=[],
                country_codes=[],
            ),
        ),
        (
            """
            slug = "gammalan"
            title = "GammaLAN"
            country_codes = ["ca", "us"]
            """,
            Series(
                slug='gammalan',
                title='GammaLAN',
                alternative_titles=[],
                country_codes=['ca', 'us'],
            ),
        ),
        (
            """
            slug = "deltalan"
            title = "DeltaLAN"
            alternative_titles = ["Δ LAN", "Δέλτα LAN"]
            country_codes = ["au"]

            [links.website]
            url = "https://www.deltalan.example/"
            """,
            Series(
                slug='deltalan',
                title='DeltaLAN',
                alternative_titles=['Δ LAN', 'Δέλτα LAN'],
                country_codes=['au'],
                links=SeriesLinks(
                    website=Resource(
                        url='https://www.deltalan.example/',
                        offline=False,
                    ),
                ),
            ),
        ),
        (
            """
            slug = "epsilan"
            title = "EpsiLAN"
            country_codes = ["gr"]

            [links.website]
            url = "https://www.epsilan.example/"
            offline = true
            """,
            Series(
                slug='epsilan',
                title='EpsiLAN',
                country_codes=['gr'],
                links=SeriesLinks(
                    website=Resource(
                        url='https://www.epsilan.example/',
                        offline=True,
                    ),
                ),
            ),
        ),
    ],
)
def test_deserialize_series_from_toml(toml: str, expected: Series):
    assert deserialize_series_from_toml(toml) == expected



"""
lanpartydb.serialization
~~~~~~~~~~~~~~~~~~~~~~~~

Object serialization to TOML

:Copyright: 2024-2025 Jochen Kupperschmidt
:License: MIT
"""

import dataclasses
from typing import Any

from lanpartydb.models import Party, Series
import tomlkit


# series


def serialize_series_list_to_toml(series_list: list[Series]) -> str:
    """Serialize list of series to TOML document."""
    aot = tomlkit.aot()

    for series in series_list:
        series_dict = _series_to_sparse_dict(series)
        aot.append(tomlkit.item(series_dict))

    doc = tomlkit.document()
    doc.append('series', aot)

    return _write_toml(doc)


def serialize_series_to_toml(series: Series) -> str:
    """Serialize series to TOML document."""
    series_dict = _series_to_sparse_dict(series)

    return _write_toml(series_dict)


def _series_to_sparse_dict(series: Series) -> dict[str, Any]:
    data = dataclasses.asdict(series)
    _remove_default_values(data)
    return data


# party


def serialize_party_to_toml(party: Party) -> str:
    """Serialize party to TOML document."""
    party_dict = _party_to_sparse_dict(party)

    location = party_dict.get('location', None)
    if location is not None:
        _convert_decimal_to_float(location, 'latitude')
        _convert_decimal_to_float(location, 'longitude')

    return _write_toml(party_dict)


def _party_to_sparse_dict(party: Party) -> dict[str, Any]:
    data = dataclasses.asdict(party)
    _remove_default_values(data)
    return data


def _convert_decimal_to_float(d: dict[str, Any], key: str) -> None:
    value = d.get(key)
    if value is not None:
        d[key] = float(value)


# util


def _write_toml(d: dict[str, Any]) -> str:
    return tomlkit.dumps(d)


def _remove_default_values(d: dict[str, Any]) -> dict[str, Any]:
    """Remove `None`, `False`, and `[]` values from first level of
    dictionary.
    """
    for k, v in list(d.items()):
        if (v is None) or (v is False) or (v == []):
            del d[k]
        elif isinstance(v, dict):
            _remove_default_values(v)

    return d
